import stk

from .selector import Selector


class AllSelector(Selector):
    """Select all atom ids."""

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple[int, ...]:
        return tuple(i.get_id() for i in molecule.get_atoms())


class AllNonHSelector(Selector):
    """Select all atom ids."""

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple[int, ...]:
        return tuple(
            i.get_id()
            for i in molecule.get_atoms()
            if i.get_atomic_number() != 1
        )



import numpy as np
import stk

from bbprep._internal.ensemble.ensemble import Conformer, Ensemble
from bbprep._internal.selectors.binders import BindersSelector
from bbprep._internal.selectors.notplacers import NotPlacersSelector
from bbprep._internal.selectors.selector import NullSelector

from .process import Process
from .utilities import angle_between


class DitopicFitter(Process):
    """Get the conformer best for binding as ditopic linker."""

    def __init__(self, ensemble: Ensemble) -> None:
        """Initialise the process."""
        self._data: dict[str, float]
        self._ensemble = ensemble
        self._selector = NullSelector()
        self._data = {}

    def _run_process(
        self,
        conformer: Conformer,
        conformer_id: int,
    ) -> float:
        key = stk.Smiles().get_key(conformer.molecule) + f"__{conformer_id}"
        if key in self._data:
            return self._data[key]

        expected = 2
        if conformer.molecule.get_num_functional_groups() != expected:
            msg = (
                f"Found {conformer.molecule.get_num_functional_groups()}"
                " functional groups, not 2."
            )
            raise RuntimeError(msg)

        notplacers = NotPlacersSelector().yield_stepwise(conformer.molecule)
        binders = BindersSelector().yield_stepwise(conformer.molecule)

        notplacers_centroids = [
            conformer.molecule.get_centroid(pl) for pl in notplacers
        ]

        binders_centroids = [
            conformer.molecule.get_centroid(pl) for pl in binders
        ]

        vectors = []
        for i, j in zip(notplacers_centroids, binders_centroids, strict=False):
            vectors.append((j - i) / np.linalg.norm(j - i))

        value = angle_between(*vectors)
        self._save_to_data(conformer, conformer_id, value)
        return self._data[key]



import itertools as it

import stk
import stko
from rdkit.Chem import AllChem, rdMolTransforms

from bbprep._internal.ensemble.ensemble import Conformer, Ensemble

from .generator import Generator
from .targets import TorsionRange


class TorsionScanner(Generator):
    """Generate conformers by scanning target torsions."""

    def __init__(
        self,
        target_torsions: TorsionRange | tuple[TorsionRange],
    ) -> None:
        """Initialise generator."""
        if not isinstance(target_torsions, tuple):
            self._target_torsions = (target_torsions,)
        else:
            self._target_torsions = target_torsions

    def generate_conformers(self, molecule: stk.BuildingBlock) -> Ensemble:
        # Optimise the initial ligand structure.
        molecule = stko.MMFF(  # type:ignore[assignment]
            ignore_inter_interactions=False
        ).optimize(
            mol=molecule,
        )
        ensemble = Ensemble(base_molecule=molecule)
        rdkit_molecule = molecule.to_rdkit_mol()
        AllChem.SanitizeMol(rdkit_molecule)
        rdkit_properties = AllChem.MMFFGetMoleculeProperties(
            rdkit_molecule, mmffVariant="MMFF94s"
        )

        matched_torsions = {}
        atoms_to_be_constrained = set()
        for target in self._target_torsions:
            matches = rdkit_molecule.GetSubstructMatches(
                query=AllChem.MolFromSmarts(target.smarts),
            )

            for match in matches:
                if len(match) != target.expected_num_atoms:
                    msg = (
                        f"{len(match)} not as expected ("
                        f"{target.expected_num_atoms})"
                    )
                    raise RuntimeError(msg)

                if not any(i in atoms_to_be_constrained for i in match):
                    for i in match:
                        atoms_to_be_constrained.add(i)

                    initial_torsion = rdMolTransforms.GetDihedralDeg(
                        rdkit_molecule.GetConformer(0),
                        match[target.scanned_ids[0]],
                        match[target.scanned_ids[1]],
                        match[target.scanned_ids[2]],
                        match[target.scanned_ids[3]],
                    )
                    key = tuple(match[i] for i in target.scanned_ids)
                    matched_torsions[key] = [
                        round(initial_torsion + angle, 2)
                        for angle in target.scanned_range
                    ]

        test_molecule = molecule.clone()
        keys, values = zip(*matched_torsions.items(), strict=False)
        permutations_dicts = [
            dict(zip(keys, v, strict=False)) for v in it.product(*values)
        ]
        for cid, permutation in enumerate(permutations_dicts):
            rdkit_molecule = test_molecule.to_rdkit_mol()
            AllChem.SanitizeMol(rdkit_molecule)
            rdkit_properties = AllChem.MMFFGetMoleculeProperties(
                rdkit_molecule
            )
            ff = AllChem.MMFFGetMoleculeForceField(
                rdkit_molecule,
                rdkit_properties,
            )
            for torsion in permutation:
                actual_angle = permutation[torsion]

                ff.MMFFAddTorsionConstraint(
                    torsion[0],
                    torsion[1],
                    torsion[2],
                    torsion[3],
                    False,  # noqa: FBT003
                    actual_angle - 0.1,
                    actual_angle + 0.1,
                    100.0,
                )

            ff.Minimize(maxIts=500)
            pos_mat = rdkit_molecule.GetConformer(-1).GetPositions()
            test_molecule = molecule.with_position_matrix(pos_mat)
            ensemble.add_conformer(
                conformer=Conformer(
                    molecule=test_molecule,
                    conformer_id=cid,
                    source="torsionscan",
                    permutation=permutation,
                ),
            )

        return ensemble



import numpy as np
import stk

from bbprep._internal.ensemble.ensemble import Conformer

from .process import Process
from .utilities import angle_between


class MinimiseAngle(Process):
    """Get the molecule with the min/max of a target angle."""

    def _run_process(
        self,
        conformer: Conformer,
        conformer_id: int,
    ) -> float:
        key = stk.Smiles().get_key(conformer.molecule) + f"__{conformer_id}"
        if key in self._data:
            return self._data[key]

        atom_positions = self._selector.get_atomic_positions(
            conformer.molecule
        )

        expected = 3
        if len(atom_positions) != expected:
            msg = f"Selector found {len(atom_positions)} atoms, not 3"
            raise RuntimeError(msg)

        vectors = (
            atom_positions[0] - atom_positions[1],
            atom_positions[2] - atom_positions[1],
        )
        value = np.degrees(angle_between(*vectors))
        self._save_to_data(conformer, conformer_id, value)
        return self._data[key]



from collections import abc

import numpy as np
import numpy.typing as npt
import stk


class Selector:
    """Select atom ids in stk molecules by deleters."""

    def __init__(self) -> None:
        """Initialise Selector."""

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple[int, ...]:
        raise NotImplementedError

    def yield_stepwise(
        self,
        molecule: stk.BuildingBlock,
    ) -> abc.Iterator[tuple[int, ...]]:
        raise NotImplementedError

    def get_atomic_positions(
        self,
        molecule: stk.BuildingBlock,
    ) -> tuple[npt.NDArray[np.float64], ...]:
        return tuple(
            molecule.get_atomic_positions(atom_ids=self.select_atoms(molecule))
        )


class NullSelector(Selector):
    """Selecter that does nothing."""

    def select_atoms(
        self,
        molecule: stk.BuildingBlock,  # noqa: ARG002
    ) -> tuple[int, ...]:
        return ()



import pathlib

import numpy as np
from bbprep import TargetTorsion

from .case_data import CaseData


def test_targettorsion(molecule: CaseData) -> None:
    """Test :class:`TargetTorsion`.

    Parameters:

        molecule:
            The molecule.

    Returns:
        None : :class:`NoneType`

    """
    ensemble = molecule.generator.generate_conformers(molecule.molecule)

    process = TargetTorsion(
        ensemble=ensemble,
        selector=molecule.selector,
        target_value=molecule.target_value,
    )

    best_molecule = process.get_best()
    path = pathlib.Path(__file__).parent
    best_molecule.molecule.write(path / f"torsion_{molecule.name}_best.mol")

    all_scores = process.get_all_scores()
    print(all_scores)
    print(
        process.get_best_id(),
        process.calculate_value(
            best_molecule,
            molecule.best_id,
        ),
    )

    assert molecule.best_id == process.get_best_id()
    assert min(all_scores) == process.calculate_score(
        best_molecule, molecule.best_id
    )
    assert np.isclose(
        molecule.best_value,
        process.calculate_value(
            best_molecule,
            molecule.best_id,
        ),
        rtol=0,
        atol=5e-1,
    )



"""selectors package."""

from bbprep._internal.selectors.all_selector import (
    AllNonHSelector,
    AllSelector,
)
from bbprep._internal.selectors.binders import BindersSelector
from bbprep._internal.selectors.by_id import ByIdSelector
from bbprep._internal.selectors.by_smarts import BySmartsSelector
from bbprep._internal.selectors.deleters import DeletersSelector
from bbprep._internal.selectors.notplacers import NotPlacersSelector
from bbprep._internal.selectors.selector import NullSelector, Selector
from bbprep._internal.selectors.xcomx import XCOMXSelector

__all__ = [
    "Selector",
    "NullSelector",
    "AllSelector",
    "AllNonHSelector",
    "BindersSelector",
    "ByIdSelector",
    "BySmartsSelector",
    "DeletersSelector",
    "NotPlacersSelector",
    "XCOMXSelector",
]



import stk
import stko

from bbprep._internal.ensemble.ensemble import Conformer

from .process import Process


class Planarfy(Process):
    """Get the most planar molecule in an ensemble."""

    def _run_process(
        self,
        conformer: Conformer,
        conformer_id: int,
    ) -> float:
        key = stk.Smiles().get_key(conformer.molecule) + f"__{conformer_id}"
        if key in self._data:
            return self._data[key]

        pc = stko.PlanarityCalculator()
        pc_results = pc.get_results(
            conformer.molecule,
            plane_atom_ids=self._selector.select_atoms(conformer.molecule),
            deviation_atom_ids=self._selector.select_atoms(conformer.molecule),
        )
        value = pc_results.get_planarity_parameter()
        self._save_to_data(conformer, conformer_id, value)
        return self._data[key]



"""generators package."""

from bbprep._internal.generators.etkdg import ETKDG
from bbprep._internal.generators.generator import Generator
from bbprep._internal.generators.geometry_scanner import GeometryScanner
from bbprep._internal.generators.scanner_by_selector import (
    SelectorDistanceScanner,
)
from bbprep._internal.generators.targets import (
    AngleRange,
    BondRange,
    TorsionRange,
)
from bbprep._internal.generators.torsion_scanner import TorsionScanner

__all__ = [
    "Generator",
    "ETKDG",
    "TorsionScanner",
    "AngleRange",
    "BondRange",
    "TorsionRange",
    "GeometryScanner",
    "SelectorDistanceScanner",
]



import stk
from rdkit.Chem import AllChem

from bbprep._internal.ensemble.ensemble import Conformer, Ensemble

from .generator import Generator


class ETKDG(Generator):
    """Generate conformers as stk molecules with :func:`rdkit.ETKDGv2()`."""

    def __init__(self, num_confs: int) -> None:
        """Initialise ETKDG generator.

        `v3`:
            New version from DOI: 10.1021/acs.jcim.0c00025
            with improved handling of macrocycles.

        """
        self._num_confs = num_confs

    def generate_conformers(
        self,
        molecule: stk.BuildingBlock,
    ) -> Ensemble:
        rdkit_molecule = molecule.to_rdkit_mol()
        rdkit_molecule.RemoveAllConformers()

        params = AllChem.ETKDGv3()
        params.randomSeed = 1000
        cids = AllChem.EmbedMultipleConfs(
            mol=rdkit_molecule,
            numConfs=self._num_confs,
            params=params,
        )

        ensemble = Ensemble(base_molecule=molecule)
        for cid in cids:
            pos_mat = rdkit_molecule.GetConformer(id=cid).GetPositions()
            ensemble.add_conformer(
                conformer=Conformer(
                    molecule=molecule.with_position_matrix(pos_mat),
                    conformer_id=cid,
                    source="etkdg-v3",
                    permutation=None,
                ),
            )

        return ensemble



from collections import abc
from dataclasses import dataclass


@dataclass
class BondRange:
    smarts: str
    expected_num_atoms: int
    scanned_ids: tuple[int, int]
    scanned_range: abc.Iterable[float]


@dataclass
class AngleRange:
    smarts: str
    expected_num_atoms: int
    scanned_ids: tuple[int, int, int]
    scanned_range: abc.Iterable[float]


@dataclass
class TorsionRange:
    smarts: str
    expected_num_atoms: int
    scanned_ids: tuple[int, int, int, int]
    scanned_range: abc.Iterable[float]



import pathlib

import stk
from bbprep import FurthestFGs

from .case_data import CaseData


def test_furtherfgs(molecule: CaseData) -> None:
    """Test :class:`FurthestFGs`.

    Parameters:

        molecule:
            The molecule to modify.

    Returns:
        None : :class:`NoneType`

    """
    modified = FurthestFGs().modify(
        building_block=molecule.molecule,
        desired_functional_groups=molecule.desired_functional_groups,
    )
    modified_fgs = set(modified.get_functional_groups())
    original_fgs = tuple(molecule.molecule.get_functional_groups())
    fg_positions = []
    for idx, fg in enumerate(original_fgs):
        if fg in modified_fgs:
            assert idx in molecule.furthest_ids
            fg_positions.append(
                molecule.molecule.get_centroid(atom_ids=fg.get_placer_ids())
            )

    xyz_string = stk.XyzWriter().to_string(molecule.molecule).split("\n")
    xyz_string[0] = (
        f"{int(xyz_string[0]) + molecule.desired_functional_groups}"
    )
    for fgpos in fg_positions:
        xyz_string.append(
            f"He {round(fgpos[0], 2)} {round(fgpos[1], 2)} "
            f"{round(fgpos[2], 2)}"
        )

    path = pathlib.Path(__file__).parent
    with open(  # noqa: PTH123
        path / f"dist_{molecule.name}_fur.xyz", "w"
    ) as f:
        f.write("\n".join(xyz_string))



import pathlib

import numpy as np
import numpy.typing as npt
import stk
from stk._internal.topology_graphs.utilities import _FunctionalGroupSorter
from stk._internal.utilities.utilities import (
    get_acute_vector,
    get_plane_normal,
)

from .modifier import Modifier


class PanelBuildingBlock(stk.BuildingBlock):
    def get_bonders(self) -> dict[int, int]:
        return {
            i: next(iter(fg.get_bonder_ids()))  # type: ignore[attr-defined]
            for i, fg in enumerate(self.get_functional_groups())
        }

    def show_long_axis(
        self,
        long_axis: npt.NDArray[np.float64],
        path: pathlib.Path,
    ) -> None:
        string = stk.XyzWriter().to_string(self)

        # Long axis is along y direction.
        x_pos = [0, long_axis[0]]
        y_pos = [0, long_axis[1]]
        z_pos = [0, long_axis[2]]

        for x, y, z in zip(x_pos, y_pos, z_pos, strict=False):
            string += f"Ar {x} {y} {z}\n"

        stringlist = string.split("\n")
        stringlist[0] = str(int(stringlist[0]) + len(x_pos))
        string = "\n".join(stringlist)

        with path.open("w") as f:
            f.write(string)

    def get_long_axis(self) -> npt.NDArray[np.float64]:
        atom_ids = self.get_bonders()
        expected = 4
        if len(atom_ids) != expected:
            msg = f"{self} has too many functional groups."
            raise RuntimeError(msg)

        sorted_fg_ids = list(_FunctionalGroupSorter(self).get_items())

        max_pair_distance = 0.0
        for i, fg_id in enumerate(sorted_fg_ids):
            id1_centroid = self.get_centroid(
                atom_ids=(atom_ids[fg_id]),
            )
            if i != 3:  # noqa: PLR2004
                id2_centroid = self.get_centroid(
                    atom_ids=(atom_ids[sorted_fg_ids[i + 1]]),
                )
            else:
                id2_centroid = self.get_centroid(
                    atom_ids=(atom_ids[sorted_fg_ids[0]]),
                )
            pair_axis = id1_centroid - id2_centroid
            pair_distance = float(np.linalg.norm(pair_axis))
            if pair_distance > max_pair_distance:
                long_axis = pair_axis.copy()
                max_pair_distance = pair_distance

        return long_axis

    def get_concave_direction(self) -> npt.NDArray[np.float64]:
        atom_ids = self.get_bonders()
        expected = 4
        if len(atom_ids) != expected:
            msg = f"{self} has too many functional groups."
            raise RuntimeError(msg)

        sorted_fg_ids = list(_FunctionalGroupSorter(self).get_items())
        pair_list = list(
            zip(
                sorted_fg_ids,
                sorted_fg_ids[1:] + sorted_fg_ids[:1],
                strict=False,
            )
        )
        max_pair_distance = 0.0
        for fg_id1, fg_id2 in pair_list:
            id1_centroid = self.get_centroid(
                atom_ids=(atom_ids[fg_id1]),
            )
            id2_centroid = self.get_centroid(
                atom_ids=(atom_ids[fg_id2]),
            )
            pair_axis = id1_centroid - id2_centroid
            pair_distance = float(np.linalg.norm(pair_axis))
            if pair_distance > max_pair_distance:
                max_pair_distance = pair_distance
                long_axis_centroid = np.divide(id1_centroid + id2_centroid, 2)
                for fg_id3, fg_id4 in pair_list:
                    if fg_id3 not in (fg_id1, fg_id2) and fg_id4 not in (
                        fg_id1,
                        fg_id2,
                    ):
                        id3_centroid = self.get_centroid(
                            atom_ids=(atom_ids[fg_id3]),
                        )
                        id4_centroid = self.get_centroid(
                            atom_ids=(atom_ids[fg_id4]),
                        )

                        short_axis_centroid = np.divide(
                            id4_centroid + id3_centroid, 2
                        )

        return short_axis_centroid - long_axis_centroid


class ReorientPanel(Modifier):
    def reassign_functional_groups(
        self,
        building_block: PanelBuildingBlock,
    ) -> PanelBuildingBlock:
        # Set functional group ordering based on long axis.
        fg_centroids = tuple(
            building_block.get_centroid(
                atom_ids=fg.get_placer_ids(),
            )
            for fg in building_block.get_functional_groups()
        )
        plus_minus_fg_id = next(
            i
            for i, cent in enumerate(fg_centroids)
            if cent[0] > 0 and cent[1] < 0
        )
        fg1_id = plus_minus_fg_id
        fg2_id, fg3_id, fg4_id = tuple(
            i
            for i in range(building_block.get_num_functional_groups())
            if i != fg1_id
        )
        new_fgs = tuple(building_block.get_functional_groups())
        return building_block.with_functional_groups(
            functional_groups=(
                new_fgs[fg1_id],
                new_fgs[fg2_id],
                new_fgs[fg3_id],
                new_fgs[fg4_id],
            )
        )


class ReorientC2Panel(ReorientPanel):
    """Modify a building block."""

    def modify(self, building_block: stk.BuildingBlock) -> PanelBuildingBlock:
        building_block = PanelBuildingBlock.init_from_molecule(
            molecule=building_block,
            functional_groups=building_block.get_functional_groups(),
        )
        expected = 4
        if building_block.get_num_functional_groups() != expected:
            msg = f"{building_block} does not have 4 functional groups."
            raise RuntimeError(msg)

        target_coords = (
            np.array(
                (
                    np.array([1, 1, 0]),
                    np.array([1, -1, 0]),
                    np.array([-1, -1, 0]),
                    np.array([-1, 1, 0]),
                )
            )
            * building_block.get_maximum_diameter()
            / 2
        )
        centroid_pos = np.array([0, 0, 0])
        building_block = building_block.with_centroid(
            position=centroid_pos,
            atom_ids=building_block.get_placer_ids(),
        )

        edge_centroid = sum(target_coords) / len(target_coords)
        edge_normal = get_acute_vector(
            reference=edge_centroid,
            vector=get_plane_normal(
                points=np.array(target_coords),
            ),
        )

        fg_bonder_centroid = building_block.get_centroid(
            atom_ids=next(
                building_block.get_functional_groups()
            ).get_placer_ids(),
        )
        edge_position = target_coords[0]
        building_block = building_block.with_rotation_to_minimize_angle(
            start=fg_bonder_centroid - centroid_pos,
            target=edge_position - edge_centroid,
            axis=edge_normal,
            origin=centroid_pos,
        )

        # Flatten wrt to xy plane.
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        normal = building_block.get_plane_normal(
            atom_ids=building_block.get_placer_ids(),
        )
        normal = get_acute_vector(
            reference=core_centroid - centroid_pos,
            vector=normal,
        )
        building_block = building_block.with_rotation_between_vectors(
            start=normal,
            target=np.array([0, 0, 1]),
            origin=centroid_pos,
        )

        # Align long axis of molecule (defined by deleter atoms) with
        # y axis.
        long_axis_vector = building_block.get_long_axis()
        edge_centroid = sum(target_coords) / len(target_coords)
        edge_normal = get_acute_vector(
            reference=edge_centroid,
            vector=get_plane_normal(
                points=np.array(target_coords),
            ),
        )
        building_block = building_block.with_rotation_to_minimize_angle(
            start=long_axis_vector,
            target=np.array([1, 0, 0]),
            axis=edge_normal,
            origin=centroid_pos,
        )

        return self.reassign_functional_groups(building_block)


class ReorientC1Panel(ReorientPanel):
    """Modify a building block."""

    def modify(self, building_block: stk.BuildingBlock) -> PanelBuildingBlock:
        building_block = PanelBuildingBlock.init_from_molecule(
            molecule=building_block,
            functional_groups=building_block.get_functional_groups(),
        )
        expected = 4
        if building_block.get_num_functional_groups() != expected:
            msg = f"{building_block} does not have 4 functional groups."
            raise RuntimeError(msg)

        target_coords = (
            np.array(
                (
                    np.array([1, 1, 0]),
                    np.array([1, -1, 0]),
                    np.array([-1, -1, 0]),
                    np.array([-1, 1, 0]),
                )
            )
            * building_block.get_maximum_diameter()
            / 2
        )
        centroid_pos = np.array([0, 0, 0])
        edge_centroid = sum(target_coords) / len(target_coords)

        building_block = building_block.with_centroid(
            position=centroid_pos,
            atom_ids=building_block.get_placer_ids(),
        )

        edge_normal = get_acute_vector(
            reference=edge_centroid,
            vector=get_plane_normal(
                points=np.array(target_coords),
            ),
        )

        fg_bonder_centroid = building_block.get_centroid(
            atom_ids=next(
                building_block.get_functional_groups()
            ).get_placer_ids(),
        )
        edge_position = target_coords[0]
        building_block = building_block.with_rotation_to_minimize_angle(
            start=fg_bonder_centroid - centroid_pos,
            target=edge_position - edge_centroid,
            axis=edge_normal,
            origin=centroid_pos,
        )

        # Flatten wrt to xy plane.
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        normal = building_block.get_plane_normal(
            atom_ids=building_block.get_placer_ids(),
        )
        normal = get_acute_vector(
            reference=core_centroid - centroid_pos,
            vector=normal,
        )
        building_block = building_block.with_rotation_between_vectors(
            start=normal,
            target=np.array([0, 0, 1]),
            origin=centroid_pos,
        )

        # Align long axis of molecule with the x axis.
        building_block = building_block.with_rotation_to_minimize_angle(
            start=building_block.get_long_axis(),
            target=np.array([1, 0, 0]),
            axis=edge_normal,
            origin=centroid_pos,
        )
        # Align concave axis of molecule with the y axis.
        building_block = building_block.with_rotation_to_minimize_angle(
            start=building_block.get_concave_direction(),
            target=np.array([0, 1, 0]),
            axis=edge_normal,
            origin=centroid_pos,
        )

        return self.reassign_functional_groups(building_block)



import itertools as it

import stk
import stko
from rdkit.Chem import AllChem, rdMolTransforms

from bbprep._internal.ensemble.ensemble import Conformer, Ensemble

from .generator import Generator
from .targets import AngleRange, BondRange, TorsionRange


class GeometryScanner(Generator):
    """Generate conformers by scanning multiple settable geometries."""

    def __init__(
        self,
        target_ranges: BondRange
        | AngleRange
        | TorsionRange
        | tuple[BondRange | AngleRange | TorsionRange],
    ) -> None:
        """Initialise generator."""
        if not isinstance(target_ranges, tuple):
            self._target_ranges = (target_ranges,)
        else:
            self._target_ranges = target_ranges

    def generate_conformers(  # noqa: C901, PLR0912
        self,
        molecule: stk.BuildingBlock,
    ) -> Ensemble:
        # Optimise the initial ligand structure.
        molecule = stko.MMFF(  # type:ignore[assignment]
            ignore_inter_interactions=False
        ).optimize(
            mol=molecule,
        )
        ensemble = Ensemble(base_molecule=molecule)
        rdkit_molecule = molecule.to_rdkit_mol()
        AllChem.SanitizeMol(rdkit_molecule)
        rdkit_properties = AllChem.MMFFGetMoleculeProperties(
            rdkit_molecule, mmffVariant="MMFF94s"
        )

        matched_changes = {}
        atoms_to_be_constrained_bonds = set()
        atoms_to_be_constrained_angles = set()
        atoms_to_be_constrained_torsions = set()
        for target in self._target_ranges:
            matches = rdkit_molecule.GetSubstructMatches(
                query=AllChem.MolFromSmarts(target.smarts),
            )

            for match in matches:
                if len(match) != target.expected_num_atoms:
                    msg = (
                        f"{len(match)} not as expected ("
                        f"{target.expected_num_atoms})"
                    )
                    raise RuntimeError(msg)
                if isinstance(target, AngleRange) and not any(
                    i in atoms_to_be_constrained_angles for i in match
                ):
                    for i in match:
                        atoms_to_be_constrained_angles.add(i)
                    initial_value = rdMolTransforms.GetAngleDeg(
                        rdkit_molecule.GetConformer(0),
                        match[target.scanned_ids[0]],
                        match[target.scanned_ids[1]],
                        match[target.scanned_ids[2]],
                    )
                    key = tuple(match[i] for i in target.scanned_ids)
                    matched_changes[key] = [
                        round(initial_value + test, 2)
                        for test in target.scanned_range
                    ]

                elif isinstance(target, TorsionRange) and not any(
                    i in atoms_to_be_constrained_torsions for i in match
                ):
                    for i in match:
                        atoms_to_be_constrained_torsions.add(i)
                    initial_value = rdMolTransforms.GetDihedralDeg(
                        rdkit_molecule.GetConformer(0),
                        match[target.scanned_ids[0]],
                        match[target.scanned_ids[1]],
                        match[target.scanned_ids[2]],
                        match[target.scanned_ids[3]],
                    )
                    key = tuple(match[i] for i in target.scanned_ids)
                    matched_changes[key] = [
                        round(initial_value + test, 2)
                        for test in target.scanned_range
                    ]

                elif isinstance(target, BondRange) and not any(
                    i in atoms_to_be_constrained_bonds for i in match
                ):
                    for i in match:
                        atoms_to_be_constrained_bonds.add(i)
                    initial_value = rdMolTransforms.GetBondLength(
                        rdkit_molecule.GetConformer(0),
                        match[target.scanned_ids[0]],
                        match[target.scanned_ids[1]],
                    )

                    key = tuple(match[i] for i in target.scanned_ids)
                    matched_changes[key] = [
                        round(initial_value + test, 2)
                        for test in target.scanned_range
                    ]

        test_molecule = molecule.clone()
        keys, values = zip(*matched_changes.items(), strict=False)
        permutations_dicts = [
            dict(zip(keys, v, strict=False)) for v in it.product(*values)
        ]

        for cid, permutation in enumerate(permutations_dicts):
            rdkit_molecule = test_molecule.to_rdkit_mol()
            AllChem.SanitizeMol(rdkit_molecule)
            rdkit_properties = AllChem.MMFFGetMoleculeProperties(
                rdkit_molecule
            )
            ff = AllChem.MMFFGetMoleculeForceField(
                rdkit_molecule,
                rdkit_properties,
            )
            for change in permutation:
                actual_value = permutation[change]

                if len(change) == 2:  # noqa: PLR2004
                    # Add constraint.
                    ff.MMFFAddDistanceConstraint(
                        change[0],
                        change[1],
                        False,  # noqa: FBT003
                        actual_value - 0.01,
                        actual_value + 0.01,
                        1.0e5,
                    )

                elif len(change) == 3:  # noqa: PLR2004
                    # Add constraint.
                    ff.MMFFAddAngleConstraint(
                        change[0],
                        change[1],
                        change[2],
                        False,  # noqa: FBT003
                        actual_value - 0.1,
                        actual_value + 0.1,
                        100.0,
                    )

                elif len(change) == 4:  # noqa: PLR2004
                    # Add constraint.
                    ff.MMFFAddTorsionConstraint(
                        change[0],
                        change[1],
                        change[2],
                        change[3],
                        False,  # noqa: FBT003
                        actual_value - 0.1,
                        actual_value + 0.1,
                        100.0,
                    )

            ff.Minimize(maxIts=500)
            pos_mat = rdkit_molecule.GetConformer(-1).GetPositions()
            test_molecule = molecule.with_position_matrix(pos_mat)
            ensemble.add_conformer(
                conformer=Conformer(
                    molecule=test_molecule,
                    conformer_id=cid,
                    source="geomscan",
                    permutation=permutation,
                ),
            )

        return ensemble



from collections import abc

import stk

from .selector import Selector


class DeletersSelector(Selector):
    """Select atom ids in stk molecules by deleters."""

    def __init__(self) -> None:
        """Initialise Selector."""

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple[int, ...]:
        if molecule.get_num_functional_groups() == 0:
            msg = "Molecule has zero functional groups."
            raise ValueError(msg)

        atoms = [
            id_
            for fg in molecule.get_functional_groups()
            for id_ in fg.get_deleter_ids()  # type: ignore[attr-defined]
        ]

        return tuple(atoms)

    def yield_stepwise(
        self,
        molecule: stk.BuildingBlock,
    ) -> abc.Iterator[tuple[int, ...]]:
        if molecule.get_num_functional_groups() == 0:
            msg = "Molecule has zero functional groups."
            raise ValueError(msg)

        for fg in molecule.get_functional_groups():
            atoms = list(fg.get_deleter_ids())  # type: ignore[attr-defined]
            yield tuple(atoms)



"""bbprep package."""

from bbprep import generators, selectors
from bbprep._internal.ensemble.ensemble import Conformer, Ensemble
from bbprep._internal.modifiers.distanced_functional_groups import (
    ClosestFGs,
    FurthestFGs,
)
from bbprep._internal.modifiers.modifier import Modifier
from bbprep._internal.modifiers.random_functional_groups import RandomFGs
from bbprep._internal.modifiers.reorient_panel import (
    PanelBuildingBlock,
    ReorientC1Panel,
    ReorientC2Panel,
    ReorientPanel,
)
from bbprep._internal.processes.angle import MinimiseAngle
from bbprep._internal.processes.ditopicfitter import DitopicFitter
from bbprep._internal.processes.planarfy import Planarfy
from bbprep._internal.processes.process import Process, TargetProcess
from bbprep._internal.processes.torsion import TargetTorsion

__all__ = [
    "generators",
    "selectors",
    "Ensemble",
    "DitopicFitter",
    "PanelBuildingBlock",
    "ReorientPanel",
    "ReorientC2Panel",
    "ReorientC1Panel",
    "Conformer",
    "Process",
    "TargetProcess",
    "Planarfy",
    "MinimiseAngle",
    "TargetTorsion",
    "Modifier",
    "FurthestFGs",
    "ClosestFGs",
    "RandomFGs",
]



from collections import abc

import stk
from rdkit.Chem import AllChem

from .selector import Selector


class BySmartsSelector(Selector):
    """Select atom ids in stk molecules by smarts string."""

    def __init__(self, smarts: str, selected_indices: tuple[int, ...]) -> None:
        """Initialise Selector."""
        self._smarts = smarts
        self._selected_indices = selected_indices

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple[int, ...]:
        rdkit_mol = molecule.to_rdkit_mol()
        AllChem.SanitizeMol(rdkit_mol)
        matches = rdkit_mol.GetSubstructMatches(
            query=AllChem.MolFromSmarts(self._smarts),
        )
        atoms = []
        for match in matches:
            for idx, atom_id in enumerate(match):
                if idx in self._selected_indices:
                    atoms.append(atom_id)
        return tuple(atoms)

    def yield_stepwise(
        self,
        molecule: stk.BuildingBlock,
    ) -> abc.Iterator[tuple[int, ...]]:
        rdkit_mol = molecule.to_rdkit_mol()
        AllChem.SanitizeMol(rdkit_mol)
        matches = rdkit_mol.GetSubstructMatches(
            query=AllChem.MolFromSmarts(self._smarts),
        )
        for match in matches:
            atoms = []
            for idx, atom_id in enumerate(match):
                if idx in self._selected_indices:
                    atoms.append(atom_id)
            yield tuple(atoms)



from collections import abc

import stk

from bbprep._internal.ensemble.ensemble import Conformer, Ensemble
from bbprep._internal.selectors.selector import Selector


class Process:
    """Get the molecule with some property."""

    def __init__(self, ensemble: Ensemble, selector: Selector) -> None:
        """Initialise the process."""
        self._data: dict[str, float]
        self._ensemble = ensemble
        self._selector = selector
        self._data = {}

    def _save_to_data(
        self,
        conformer: Conformer,
        conformer_id: int,
        value: float,
    ) -> None:
        key = stk.Smiles().get_key(conformer.molecule) + f"__{conformer_id}"
        self._data[key] = value

    def _run_process(
        self,
        conformer: Conformer,
        conformer_id: int,
    ) -> float:
        raise NotImplementedError

    def get_data(self) -> dict[str, float]:
        return self._data

    def calculate_score(
        self,
        conformer: Conformer,
        conformer_id: int,
    ) -> float:
        return self._run_process(conformer, conformer_id)

    def get_all_scores(self) -> abc.Iterable[float]:
        return [
            self._run_process(conformer, conformer.conformer_id)
            for conformer in self._ensemble.yield_conformers()
        ]

    def get_minimum(self) -> Conformer:
        minimum_score = 1e24
        minimum_conformer = Conformer(
            molecule=self._ensemble.get_base_molecule().clone(),
            conformer_id=-1,
            source=None,
            permutation=None,
        )
        for conformer in self._ensemble.yield_conformers():
            score = self._run_process(conformer, conformer.conformer_id)
            if score < minimum_score:
                minimum_score = score
                minimum_conformer = Conformer(
                    molecule=conformer.molecule.clone(),
                    conformer_id=conformer.conformer_id,
                    source=conformer.source,
                    permutation=conformer.permutation,
                )
        return minimum_conformer

    def get_minimum_id(self) -> int:
        minimum_score = 1e24
        minimum_id = -1
        for conformer in self._ensemble.yield_conformers():
            score = self._run_process(conformer, conformer.conformer_id)
            if score < minimum_score:
                minimum_score = score
                minimum_id = conformer.conformer_id
        return minimum_id


class TargetProcess(Process):
    """Get the conformer with some target value."""

    def __init__(
        self,
        ensemble: Ensemble,
        selector: Selector,
        target_value: float,
    ) -> None:
        """Initialise the process."""
        self._data: dict[str, float]
        self._ensemble = ensemble
        self._selector = selector
        self._target_value = target_value
        self._data = {}

    def calculate_value(
        self,
        conformer: Conformer,
        conformer_id: int,
    ) -> float:
        return self._run_process(conformer, conformer_id)

    def calculate_score(
        self,
        conformer: Conformer,
        conformer_id: int,
    ) -> float:
        return abs(
            self._run_process(conformer, conformer_id) - self._target_value
        )

    def get_all_scores(self) -> abc.Iterable[float]:
        return [
            abs(
                self._run_process(conformer, conformer.conformer_id)
                - self._target_value
            )
            for conformer in self._ensemble.yield_conformers()
        ]

    def get_best(self) -> Conformer:
        best_score = 1e24
        best_conformer = Conformer(
            molecule=self._ensemble.get_base_molecule().clone(),
            conformer_id=-1,
            source=None,
            permutation=None,
        )
        for conformer in self._ensemble.yield_conformers():
            score = self.calculate_score(conformer, conformer.conformer_id)
            if score < best_score:
                best_score = score
                best_conformer = Conformer(
                    molecule=conformer.molecule.clone(),
                    conformer_id=conformer.conformer_id,
                    source=conformer.source,
                    permutation=conformer.permutation,
                )
        return best_conformer

    def get_best_id(self) -> int:
        best_score = 1e24
        best_id = -1
        for conformer in self._ensemble.yield_conformers():
            score = self.calculate_score(conformer, conformer.conformer_id)
            if score < best_score:
                best_score = score
                best_id = conformer.conformer_id
        return best_id



# Distributed under the terms of the MIT License.

"""Module for ensemble class.

Author: Andrew Tarzia

"""

from collections import abc
from dataclasses import dataclass

import stk


@dataclass
class Conformer:
    molecule: stk.BuildingBlock
    conformer_id: int
    source: str | None = None
    permutation: dict[tuple[int], float] | None = None


class Ensemble:
    def __init__(
        self,
        base_molecule: stk.BuildingBlock,
    ) -> None:
        self._base_molecule = base_molecule
        self._molecule_num_atoms = base_molecule.get_num_atoms()
        self._conformers: dict[int, Conformer] = {}

    def get_num_conformers(self) -> int:
        return len(self._conformers)

    def add_conformer(self, conformer: Conformer) -> None:
        conf_id = conformer.conformer_id

        if conf_id in self._conformers:
            msg = f"{conf_id} already in conformer list"
            raise RuntimeError(msg)

        self._conformers[conf_id] = conformer

    def yield_conformers(self) -> abc.Iterator[Conformer]:
        for conf_id in self._conformers:
            yield self._conformers[conf_id]

    def get_base_molecule(self) -> stk.BuildingBlock:
        return self._base_molecule

    def get_molecule_num_atoms(self) -> int:
        return self._molecule_num_atoms

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"num_confs={self.get_num_conformers()})"
        )

    def __repr__(self) -> str:
        return str(self)



import pathlib

import numpy as np
from bbprep import Planarfy

from .case_data import CaseData


def test_planarfy(molecule: CaseData) -> None:
    """Test :class:`Planarfy`.

    Parameters:

        molecule:
            The molecule to planarfy.

    Returns:
        None : :class:`NoneType`

    """
    ensemble = molecule.generator.generate_conformers(molecule.molecule)

    process = Planarfy(ensemble=ensemble, selector=molecule.selector)

    min_molecule = process.get_minimum()
    path = pathlib.Path(__file__).parent
    min_molecule.molecule.write(path / f"planar_{molecule.name}_min.mol")

    all_scores = process.get_all_scores()
    print(all_scores)
    print(
        process.get_minimum_id(),
        process.calculate_score(min_molecule, molecule.min_id),
    )

    assert np.isclose(molecule.min_value, min(all_scores), rtol=0, atol=1e-3)
    assert molecule.min_id == process.get_minimum_id()
    assert min(all_scores) == process.calculate_score(
        min_molecule, molecule.min_id
    )



import numpy as np
import pytest
import stk
from bbprep import ReorientC1Panel, ReorientC2Panel

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=C2C=C(C(=CC2=CC(=C1Br)Br)Br)Br",
                functional_groups=stk.BromoFactory(),
            ).with_rotation_about_axis(
                angle=20,
                axis=np.array((1, 2, 3)),
                origin=np.array((0, 0, 0)),
            ),
            orientmethod=ReorientC2Panel(),
            fg_reorder=(1, 0, 3, 2),
            mapping={0: 0, 1: 1, 2: 2, 3: 3},
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=C2C=C(C(=CC2=CC(=C1Br)Br)Br)Br",
                functional_groups=stk.BromoFactory(),
            ).with_rotation_about_axis(
                angle=20,
                axis=np.array((1, 2, 3)),
                origin=np.array((0, 0, 0)),
            ),
            orientmethod=ReorientC1Panel(),
            fg_reorder=(1, 0, 3, 2),
            mapping={0: 0, 1: 1, 2: 2, 3: 3},
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=C2C=C(C(=CC2=CC(=C1Br)I)Br)I",
                functional_groups=(
                    stk.BromoFactory(),
                    stk.IodoFactory(),
                ),
            ).with_rotation_about_axis(
                angle=50,
                axis=np.array((1, 2, 3)),
                origin=np.array((0, 0, 0)),
            ),
            orientmethod=ReorientC2Panel(),
            fg_reorder=(1, 0, 2, 3),
            mapping={0: 2, 1: 0, 2: 1, 3: 3},
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles=(
                    "C1=CC(=CC=C1N(C2=CC=C(C=C2)Br)C3=CC=C(C=C3)Br)N(C4"
                    "=CC=C(C=C4)Br)C5=CC=C(C=C5)Br"
                ),
                functional_groups=stk.BromoFactory(),
            ).with_rotation_about_axis(
                angle=50,
                axis=np.array((1, 2, 3)),
                origin=np.array((0, 0, 0)),
            ),
            orientmethod=ReorientC2Panel(),
            fg_reorder=(1, 0, 2, 3),
            mapping={0: 1, 1: 0, 2: 2, 3: 3},
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles=(
                    "C1=CC(=CC=C1N(C2=CC=C(C=C2)Br)C3=CC=C(S3)N(C4=CC=C"
                    "(C=C4)Br)C5=CC=C(C=C5)Br)Br"
                ),
                functional_groups=stk.BromoFactory(),
            ).with_rotation_about_axis(
                angle=80,
                axis=np.array((1, 2, 3)),
                origin=np.array((0, 0, 0)),
            ),
            orientmethod=ReorientC1Panel(),
            fg_reorder=(1, 0, 3, 2),
            mapping={0: 1, 1: 0, 2: 2, 3: 3},
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles=(
                    "BrC1C=CC(=CC=1)N(C#CI)C1=CC=C(N(C#CI)C2C=CC(=CC=2)"
                    "Br)S1"
                ),
                functional_groups=(
                    stk.BromoFactory(),
                    stk.IodoFactory(),
                ),
            ).with_rotation_about_axis(
                angle=80,
                axis=np.array((1, 2, 3)),
                origin=np.array((0, 0, 0)),
            ),
            orientmethod=ReorientC1Panel(),
            fg_reorder=(3, 0, 1, 2),
            mapping={0: 1, 1: 0, 2: 2, 3: 3},
            name=name,
        ),
    )
)
def molecule(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import itertools as it
from collections import abc

import numpy as np
import stk
import stko
from rdkit.Chem import AllChem

from bbprep._internal.ensemble.ensemble import Conformer, Ensemble
from bbprep.selectors import Selector

from .generator import Generator


class SelectorDistanceScanner(Generator):
    """Generate conformers by scanning over one selector."""

    def __init__(
        self,
        selector: Selector,
        scanned_changes: abc.Iterable[float],
    ) -> None:
        """Initialise generator."""
        self._selector = selector
        self._scanned_changes = scanned_changes

    def generate_conformers(
        self,
        molecule: stk.BuildingBlock,
    ) -> Ensemble:
        # Optimise the initial ligand structure.
        molecule = stko.MMFF(  # type:ignore[assignment]
            ignore_inter_interactions=False
        ).optimize(
            mol=molecule,
        )
        ensemble = Ensemble(base_molecule=molecule)
        rdkit_molecule = molecule.to_rdkit_mol()
        AllChem.SanitizeMol(rdkit_molecule)
        rdkit_properties = AllChem.MMFFGetMoleculeProperties(
            rdkit_molecule, mmffVariant="MMFF94s"
        )

        selected_ids = self._selector.select_atoms(molecule)
        atom_positions = self._selector.get_atomic_positions(molecule)

        initial_value = float(
            np.linalg.norm(atom_positions[1] - atom_positions[0])
        )

        key = tuple(i for i in selected_ids)
        matched_changes = {
            key: [
                round(initial_value + test, 2)
                for test in self._scanned_changes
            ]
        }

        test_molecule = molecule.clone()
        keys, values = zip(*matched_changes.items(), strict=False)
        permutations_dicts = [
            dict(zip(keys, v, strict=False)) for v in it.product(*values)
        ]

        for cid, permutation in enumerate(permutations_dicts):
            rdkit_molecule = test_molecule.to_rdkit_mol()
            AllChem.SanitizeMol(rdkit_molecule)
            rdkit_properties = AllChem.MMFFGetMoleculeProperties(
                rdkit_molecule
            )
            ff = AllChem.MMFFGetMoleculeForceField(
                rdkit_molecule,
                rdkit_properties,
            )
            for change in permutation:
                actual_value = permutation[change]

                # Add constraint.
                ff.MMFFAddDistanceConstraint(
                    change[0],
                    change[1],
                    False,  # noqa: FBT003
                    actual_value - 0.01,
                    actual_value + 0.01,
                    1.0e5,
                )

            ff.Minimize(maxIts=500)
            pos_mat = rdkit_molecule.GetConformer(-1).GetPositions()
            test_molecule = molecule.with_position_matrix(pos_mat)
            ensemble.add_conformer(
                conformer=Conformer(
                    molecule=test_molecule,
                    conformer_id=cid,
                    source="distscan",
                    permutation=permutation,
                ),
            )

        return ensemble



import numpy as np
import numpy.typing as npt


def angle_between(
    v1: npt.NDArray[np.float64],
    v2: npt.NDArray[np.float64],
    normal: npt.NDArray[np.float64] | None = None,
) -> float:
    """Returns the angle in radians between vectors 'v1' and 'v2'.

    Defined as ::

        >>> angle_between((1, 0, 0), (0, 1, 0))
        1.5707963267948966
        >>> angle_between((1, 0, 0), (1, 0, 0))
        0.0
        >>> angle_between((1, 0, 0), (-1, 0, 0))
        3.141592653589793

    https://stackoverflow.com/questions/2827393/
    angles-between-two-n-dimensional-vectors-in-python/
    13849249#13849249

    If normal is given, the angle polarity is determined using the
    cross product of the two vectors.

    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    angle = np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))

    if normal is not None:
        # Get normal vector and cross product to determine sign.
        cross = np.cross(v1_u, v2_u)
        if np.dot(normal, cross) < 0:
            angle = -angle
    return angle


def unit_vector(vector: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """Returns the unit vector of the vector.

    https://stackoverflow.com/questions/2827393/
    angles-between-two-n-dimensional-vectors-in-python/
    13849249#13849249

    """
    return vector / np.linalg.norm(vector)


def get_dihedral(
    pt1: npt.NDArray[np.float64],
    pt2: npt.NDArray[np.float64],
    pt3: npt.NDArray[np.float64],
    pt4: npt.NDArray[np.float64],
) -> float:
    """Calculate the dihedral between four points.

    Uses Praxeolitic formula --> 1 sqrt, 1 cross product
    Output in range (-pi to pi).
    From: https://stackoverflow.com/questions/20305272/
    dihedral-torsion-angle-from-four-points-in-cartesian-
    coordinates-in-python
    (new_dihedral(p))

    """
    p0 = np.asarray(pt1)
    p1 = np.asarray(pt2)
    p2 = np.asarray(pt3)
    p3 = np.asarray(pt4)

    b0 = -1.0 * (p1 - p0)
    b1 = p2 - p1
    b2 = p3 - p2

    # normalize b1 so that it does not influence magnitude of vector
    # rejections that come next
    b1 /= np.linalg.norm(b1)

    # vector rejections
    # v = projection of b0 onto plane perpendicular to b1
    #   = b0 minus component that aligns with b1
    # w = projection of b2 onto plane perpendicular to b1
    #   = b2 minus component that aligns with b1
    v = b0 - np.dot(b0, b1) * b1
    w = b2 - np.dot(b2, b1) * b1

    # angle between v and w in a plane is the torsion angle
    # v and w may not be normalized but that's fine since tan is y/x
    x = np.dot(v, w)
    y = np.dot(np.cross(b1, v), w)
    return np.degrees(np.arctan2(y, x))



import stk

from bbprep._internal.ensemble.ensemble import Conformer

from .process import TargetProcess
from .utilities import get_dihedral


class TargetTorsion(TargetProcess):
    """Get the molecule with the closest torsion to the target."""

    def _run_process(
        self,
        conformer: Conformer,
        conformer_id: int,
    ) -> float:
        key = stk.Smiles().get_key(conformer.molecule) + f"__{conformer_id}"
        if key in self._data:
            return self._data[key]

        atom_positions = self._selector.get_atomic_positions(
            conformer.molecule
        )

        expected = 4
        if len(atom_positions) != expected:
            msg = f"Selector found {len(atom_positions)} atoms, not 4"
            raise RuntimeError(msg)

        value = get_dihedral(
            pt1=atom_positions[0],
            pt2=atom_positions[1],
            pt3=atom_positions[2],
            pt4=atom_positions[3],
        )
        self._save_to_data(conformer, conformer_id, value)
        return self._data[key]



import bbprep
import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1C=CN=CC=1C1=CC=CC(C2=CC=CN=C2)=C1",
                functional_groups=stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#6]",
                    bonders=(1,),
                    deleters=(),
                ),
            ),
            generator=bbprep.generators.ETKDG(num_confs=30),
            min_value=0.0165,
            min_id=5,
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1C=CN=CC=1C1=CC=C(C2C=NC=CC=2)C=C1",
                functional_groups=stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#6]",
                    bonders=(1,),
                    deleters=(),
                ),
            ),
            generator=bbprep.generators.ETKDG(num_confs=30),
            min_value=0.9559,
            min_id=0,
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1(C=NC=CC=1)C1C=CC=C(C#CC2=CC=CC3=C2C=CN=C3)C=1",
                functional_groups=stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#6]",
                    bonders=(1,),
                    deleters=(),
                ),
            ),
            generator=bbprep.generators.ETKDG(num_confs=30),
            min_value=0.3642,
            min_id=11,
            name=name,
        ),
    )
)
def molecule(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import itertools as it

import numpy as np
import stk

from .modifier import Modifier


class DistancedFGs(Modifier):
    """Modify a building block."""

    def __init__(self) -> None:
        """Initialise the process."""
        self._reverse: bool

    def modify(  # type: ignore[override]
        self,
        building_block: stk.BuildingBlock,
        desired_functional_groups: int,
    ) -> stk.BuildingBlock:
        selected_fgs: list[stk.FunctionalGroup]
        if (
            building_block.get_num_functional_groups()
            == desired_functional_groups
        ):
            return building_block.clone()

        if (
            building_block.get_num_functional_groups()
            < desired_functional_groups
        ):
            msg = (
                f"{building_block} has less functional groups than"
                f" asked for ({desired_functional_groups})."
            )
            raise RuntimeError(msg)

        fg_centroids = [
            (
                fg,
                building_block.get_centroid(atom_ids=fg.get_placer_ids()),
            )
            for fg in building_block.get_functional_groups()
        ]
        fg_distances = sorted(
            [
                (i[0], j[0], np.linalg.norm(i[1] - j[1]))
                for i, j in it.combinations(fg_centroids, 2)
            ],
            key=lambda x: x[2],
            reverse=self._reverse,
        )

        selected_fgs = []
        for idx, _ in enumerate(fg_distances):
            if len(selected_fgs) == 0:
                selected_fgs.append(fg_distances[0][0])
                selected_fgs.append(fg_distances[0][1])
            elif fg_distances[idx][0] in set(selected_fgs):
                selected_fgs.append(fg_distances[idx][1])
            elif fg_distances[idx][1] in set(selected_fgs):
                selected_fgs.append(fg_distances[idx][0])

            if len(selected_fgs) == desired_functional_groups:
                break

        return building_block.with_functional_groups(selected_fgs)


class FurthestFGs(DistancedFGs):
    """Modify a building block."""

    def __init__(self) -> None:
        """Initialise the process."""
        self._reverse = True


class ClosestFGs(DistancedFGs):
    """Modify a building block."""

    def __init__(self) -> None:
        """Initialise the process."""
        self._reverse = False



import pathlib

import numpy as np
import stk

from .case_data import CaseData


def test_reorientpanel(molecule: CaseData) -> None:
    """Test :class:`ReorientC2Panel` and :class:`ReorientC1Panel`.

    Parameters:

        molecule:
            The molecule to modify.

    Returns:
        None : :class:`NoneType`

    """
    path = pathlib.Path(__file__).parent
    target_coords = (
        np.array(
            (
                np.array([1, 1, 0]),
                np.array([1, -1, 0]),
                np.array([-1, -1, 0]),
                np.array([-1, 1, 0]),
            )
        )
        * molecule.molecule.get_maximum_diameter()
        / 2
    )
    modified = molecule.orientmethod.modify(molecule.molecule)
    modified.write(path / f"panel_{molecule.name}_final.mol")

    fg_positions = [
        modified.get_centroid(atom_ids=fg.get_placer_ids())
        for fg in modified.get_functional_groups()
    ]

    # Write out the structure.
    xyz_string = stk.XyzWriter().to_string(molecule.molecule).split("\n")
    xyz_string += stk.XyzWriter().to_string(modified).split("\n")[2:]
    xyz_string[0] = f"{int(xyz_string[0])*2 + 4}"
    fg_distances = []
    for pos in target_coords:
        xyz_string.append(
            f"He {round(pos[0], 2)} {round(pos[1], 2)} " f"{round(pos[2], 2)}"
        )
        fg_distances.append([np.linalg.norm(i - pos) for i in fg_positions])

    with open(path / f"panel_{molecule.name}.xyz", "w") as f:  # noqa: PTH123
        f.write("\n".join(xyz_string))

    modified_fgs = list(modified.get_functional_groups())
    original_fgs = list(molecule.molecule.get_functional_groups())
    print(modified_fgs)
    print(original_fgs)
    mapping = {}
    for idx, fg in enumerate(modified_fgs):
        old_idx = original_fgs.index(fg)
        mapping[idx] = old_idx
    assert mapping == molecule.mapping

    # Test that the reassigned functional groups match expectations.
    print(fg_distances)
    assert tuple(i.index(min(i)) for i in fg_distances) == molecule.fg_reorder



from bbprep import RandomFGs

from .case_data import CaseData


def test_randomfgs(molecule: CaseData) -> None:
    """Test :class:`RandomFGs`.

    Parameters:

        molecule:
            The molecule to modify.

    Returns:
        None : :class:`NoneType`

    """
    original_fgs = tuple(molecule.molecule.get_functional_groups())
    for seed in (123, 985, 23):
        modified = RandomFGs().modify(
            building_block=molecule.molecule,
            seed=seed,
            desired_functional_groups=molecule.desired_functional_groups,
        )
        modified_fgs = set(modified.get_functional_groups())
        assert len(modified_fgs) == molecule.desired_functional_groups
        for idx, fg in enumerate(original_fgs):
            if fg in modified_fgs:
                assert idx in molecule.random_ids[seed]



import bbprep
import stk


class CaseData:
    """A test case."""

    def __init__(
        self,
        molecule: stk.Molecule,
        generator: bbprep.generators.Generator,
        selector: bbprep.selectors.Selector,
        min_id: int,
        min_value: float,
        name: str,
    ) -> None:
        self.molecule = molecule
        self.generator = generator
        self.selector = selector
        self.min_id = min_id
        self.min_value = min_value
        self.name = name



import stk

from .selector import Selector


class ByIdSelector(Selector):
    """Select atom ids in stk molecules by atom ids."""

    def __init__(self, ids: tuple[int]) -> None:
        """Initialise Selector."""
        self._ids = ids

    def select_atoms(self, molecule: stk.BuildingBlock) -> tuple[int, ...]:
        return tuple(
            i.get_id() for i in molecule.get_atoms(atom_ids=self._ids)
        )



import numpy as np
import numpy.typing as npt
import stk

from .selector import Selector


class XCOMXSelector(Selector):
    """Get binder atom positions with the molecule centroid in the middle."""

    def select_atoms(
        self,
        molecule: stk.BuildingBlock,  # noqa: ARG002
    ) -> tuple[int, ...]:
        msg = (
            "This class can only provide positions, because one "
            "position is the centroid of the molecule."
        )
        raise NotImplementedError(msg)

    def get_atomic_positions(
        self,
        molecule: stk.BuildingBlock,
    ) -> tuple[npt.NDArray[np.float64], ...]:
        if molecule.get_num_functional_groups() != 2:  # noqa: PLR2004
            msg = "Molecule does not have 2 functional groups."
            raise ValueError(msg)

        positions = [
            next(molecule.get_atomic_positions(atom_ids=id_))
            for fg in molecule.get_functional_groups()
            for id_ in fg.get_bonder_ids()  # type: ignore[attr-defined]
        ]

        positions.insert(1, molecule.get_centroid())

        return tuple(positions)



import bbprep
import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="CCC",
            ),
            generator=bbprep.generators.ETKDG(num_confs=30),
            min_value=104.235,
            selector=bbprep.selectors.AllNonHSelector(),
            min_id=6,
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=CC(=CC(=C1)C2=CN=CC=C2)C3=CN=CC=C3",
                functional_groups=stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#6]",
                    bonders=(1,),
                    deleters=(),
                ),
            ),
            generator=bbprep.generators.ETKDG(num_confs=30),
            min_value=119.237,
            selector=bbprep.selectors.XCOMXSelector(),
            min_id=26,
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=CC(=CN=C1)C2=CC=C(C=C2)C3=CN=CC=C3",
                functional_groups=stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#6]",
                    bonders=(1,),
                    deleters=(),
                ),
            ),
            min_value=149.311,
            min_id=41,
            selector=bbprep.selectors.XCOMXSelector(),
            generator=bbprep.generators.TorsionScanner(
                target_torsions=(
                    bbprep.generators.TorsionRange(
                        smarts="[#7X2]@[#6X3]@[#6X3H0]-!@[#6X3H0]@[#6X3]",
                        expected_num_atoms=5,
                        scanned_ids=(1, 2, 3, 4),
                        scanned_range=range(0, 362, 40),
                    ),
                ),
            ),
            name=name,
        ),
        # This tests the different scanning approach, should give the same
        # outcome.
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=CC(=CN=C1)C2=CC=C(C=C2)C3=CN=CC=C3",
                functional_groups=stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#6]",
                    bonders=(1,),
                    deleters=(),
                ),
            ),
            min_value=149.311,
            min_id=41,
            selector=bbprep.selectors.XCOMXSelector(),
            generator=bbprep.generators.GeometryScanner(
                target_ranges=(
                    bbprep.generators.TorsionRange(
                        smarts="[#7X2]@[#6X3]@[#6X3H0]-!@[#6X3H0]@[#6X3]",
                        expected_num_atoms=5,
                        scanned_ids=(1, 2, 3, 4),
                        scanned_range=range(0, 362, 40),
                    ),
                ),
            ),
            name=name,
        ),
    )
)
def molecule(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import stko

from .case_data import CaseData


def test_geometryscanner(molecule: CaseData) -> None:
    """Test :class:`GeometryScanner`.

    Parameters:

        molecule:
            The molecule to scan.

    Returns:
        None : :class:`NoneType`

    """
    ensemble = molecule.generator.generate_conformers(molecule.molecule)
    print(ensemble)
    assert ensemble.get_num_conformers() == molecule.num_confs
    test_energies = tuple(
        [
            stko.MMFFEnergy(ignore_inter_interactions=False).get_energy(
                conformer.molecule
            )
            for conformer in ensemble.yield_conformers()
        ]
    )

    print(
        test_energies.index(min(test_energies)),
        min(test_energies),
        molecule.min_energy,
    )
    assert (
        min(test_energies),
        test_energies.index(min(test_energies)),
    ) == molecule.min_energy
    print(
        test_energies.index(max(test_energies)),
        max(test_energies),
        molecule.max_energy,
    )
    assert (
        max(test_energies),
        test_energies.index(max(test_energies)),
    ) == molecule.max_energy
    print(test_energies[4], molecule.energy_5)
    assert (test_energies[4], 4) == molecule.energy_5



import bbprep
import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles=(
                    "C1=CC=C2C=C(C=CC2=C1)C3=CC(=CC=C3)C4=CC=CC5=CC=CC=C54"
                ),
            ),
            generator=bbprep.generators.ETKDG(num_confs=10),
            selector=bbprep.selectors.AllSelector(),
            min_value=0.8743,
            min_id=0,
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles=(
                    "C1C(Br)=C2C(C=C(C3C=CC=C(C4C5C(=CC=C(Br)C=5)C=C(Br"
                    ")C=4)C=3)C=C2)=CC=1Br"
                ),
                functional_groups=stk.BromoFactory(),
            ),
            generator=bbprep.generators.ETKDG(num_confs=10),
            selector=bbprep.selectors.BindersSelector(),
            min_value=0.2443,
            min_id=0,
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="c1ccccc1",
            ),
            generator=bbprep.generators.ETKDG(num_confs=10),
            selector=bbprep.selectors.AllSelector(),
            min_value=0.0,
            min_id=0,
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles=(
                    "C1=CC=C2C=C(C=CC2=C1)C3=CC(=CC=C3)C4=CC=CC5=CC=CC=C54"
                ),
            ),
            generator=bbprep.generators.TorsionScanner(
                target_torsions=bbprep.generators.TorsionRange(
                    smarts="[#6][#6]-!@[#6][#6]",
                    expected_num_atoms=4,
                    scanned_ids=(0, 1, 2, 3),
                    scanned_range=range(0, 362, 40),
                ),
            ),
            selector=bbprep.selectors.AllSelector(),
            min_value=0.419,
            min_id=87,
            name=name,
        ),
    )
)
def molecule(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import stk


class Modifier:
    """Modify a building block."""

    def __init__(self) -> None:
        """Initialise the process."""

    def modify(self, building_block: stk.BuildingBlock) -> stk.BuildingBlock:
        raise NotImplementedError



import stk


class CaseData:
    """A test case."""

    def __init__(
        self,
        molecule: stk.Molecule,
        desired_functional_groups: int,
        closest_ids: tuple[int, int],
        furthest_ids: tuple[int, int],
        random_ids: dict[int, tuple[int, int]],
        name: str,
    ) -> None:
        self.molecule = molecule
        self.desired_functional_groups = desired_functional_groups
        self.closest_ids = closest_ids
        self.furthest_ids = furthest_ids
        self.random_ids = random_ids
        self.name = name



import stk
from bbprep import Modifier


class CaseData:
    """A test case."""

    def __init__(
        self,
        molecule: stk.Molecule,
        orientmethod: Modifier,
        fg_reorder: tuple[int, int, int, int],
        mapping: dict[int, int],
        name: str,
    ) -> None:
        self.molecule = molecule
        self.orientmethod = orientmethod
        self.fg_reorder = fg_reorder
        self.mapping = mapping
        self.name = name



import bbprep
import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda name: CaseData(
            molecule=stk.BuildingBlock(smiles="C1=CC=NC(=C1)C2=CC=CC=N2"),
            target_value=180,
            best_value=180,
            best_id=15,
            selector=bbprep.selectors.BySmartsSelector(
                smarts="[#7][#6][#6][#7]",
                selected_indices=(0, 1, 2, 3),
            ),
            generator=bbprep.generators.ETKDG(num_confs=20),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(smiles="C[C@@H](C(=O)O)N"),
            target_value=0,
            best_value=0,
            best_id=3,
            selector=bbprep.selectors.BySmartsSelector(
                smarts="[#7][#6][#6]=[#8]",
                selected_indices=(0, 1, 2, 3),
            ),
            generator=bbprep.generators.ETKDG(num_confs=20),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(smiles="C1=CC=NC(=C1)C2=CC=CC=N2"),
            target_value=120,
            best_value=120,
            best_id=3,
            selector=bbprep.selectors.BySmartsSelector(
                smarts="[#7][#6][#6][#7]",
                selected_indices=(0, 1, 2, 3),
            ),
            generator=bbprep.generators.TorsionScanner(
                target_torsions=(
                    bbprep.generators.TorsionRange(
                        smarts="[#7][#6][#6][#7]",
                        expected_num_atoms=4,
                        scanned_ids=(0, 1, 2, 3),
                        scanned_range=range(0, 362, 40),
                    ),
                ),
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(smiles="C1=CC=NC(=C1)C2=CC=CC=N2"),
            target_value=-120,
            best_value=-120,
            best_id=6,
            selector=bbprep.selectors.BySmartsSelector(
                smarts="[#6][#7][#6][#6][#7][#6]",
                selected_indices=(1, 2, 3, 4),
            ),
            generator=bbprep.generators.TorsionScanner(
                target_torsions=bbprep.generators.TorsionRange(
                    smarts="[#7][#6][#6][#7]",
                    expected_num_atoms=4,
                    scanned_ids=(0, 1, 2, 3),
                    scanned_range=range(0, 362, 40),
                ),
            ),
            name=name,
        ),
    )
)
def molecule(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import bbprep
import stk


class CaseData:
    """A test case."""

    def __init__(
        self,
        molecule: stk.Molecule,
        generator: bbprep.generators.Generator,
        num_confs: int,
        min_energy: tuple,
        max_energy: tuple,
        energy_5: tuple,
        name: str,
    ) -> None:
        self.molecule = molecule
        self.generator = generator
        self.num_confs = num_confs
        self.min_energy = min_energy
        self.max_energy = max_energy
        self.energy_5 = energy_5
        self.name = name



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=CC(=CN=C1)C2=NC=C(C=C2)C3=CC=NC=C3",
                functional_groups=stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#6]",
                    bonders=(1,),
                    deleters=(),
                ),
            ),
            desired_functional_groups=2,
            closest_ids=(0, 1),
            furthest_ids=(0, 2),
            random_ids={123: (0, 2), 985: (1, 0), 23: (0, 2)},
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C(C(Br)Br)Br",
                functional_groups=stk.BromoFactory(),
            ),
            desired_functional_groups=2,
            closest_ids=(1, 2),
            furthest_ids=(0, 1),
            random_ids={123: (0, 2), 985: (1, 0), 23: (0, 2)},
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=CC(=C(C=C1Br)C2=CC(=CC(=C2)Br)Br)Br",
                functional_groups=stk.BromoFactory(),
            ),
            desired_functional_groups=3,
            closest_ids=(0, 2, 3),
            furthest_ids=(1, 2, 3),
            random_ids={123: (0, 2, 3), 985: (1, 0, 2), 23: (0, 1, 2)},
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1(C(C(C(C(C(C1Br)Br)Br)Br)Br)Br)Br",
                functional_groups=stk.BromoFactory(),
            ),
            desired_functional_groups=2,
            closest_ids=(0, 1),
            furthest_ids=(1, 5),
            random_ids={123: (0, 4), 985: (1, 3), 23: (0, 4)},
            name=name,
        ),
    )
)
def molecule(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import stk

from bbprep._internal.ensemble.ensemble import Ensemble


class Generator:
    """Generate an ensemble of molecules."""

    def generate_conformers(self, molecule: stk.BuildingBlock) -> Ensemble:
        raise NotImplementedError



import bbprep
import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=CC(=CC(=C1)C2=CN=CC=C2)C3=CN=CC=C3"
            ),
            generator=bbprep.generators.GeometryScanner(
                target_ranges=(
                    bbprep.generators.AngleRange(
                        smarts="[#7X2]@[#6X3]@[#6X3H0]-!@[#6X3H0]@[#6X3]",
                        expected_num_atoms=5,
                        scanned_ids=(2, 3, 4),
                        scanned_range=[-5.0, 0, 5.0, 10.0],
                    ),
                    bbprep.generators.BondRange(
                        smarts="[#7X2]@[#6X3]",
                        expected_num_atoms=2,
                        scanned_ids=(0, 1),
                        scanned_range=[-0.2, 0],
                    ),
                ),
            ),
            num_confs=64,
            min_energy=(72.8283860274666, 23),
            max_energy=(118.02644492785005, 60),
            energy_5=(111.35399028061572, 4),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(smiles="C1=CC=C(C=C1)C2=CC=CC=C2"),
            generator=bbprep.generators.GeometryScanner(
                target_ranges=(
                    bbprep.generators.BondRange(
                        smarts="[#6X3H0]-!@[#6X3H0]",
                        expected_num_atoms=2,
                        scanned_ids=(0, 1),
                        scanned_range=[-0.5, -0.3, -0.2, 0, 0.2, 0.3, 0.5],
                    ),
                ),
            ),
            num_confs=7,
            min_energy=(39.81232642917671, 3),
            max_energy=(225.68018376546095, 0),
            energy_5=(46.815865932187954, 4),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="C1=CC(=CC(=C1)Br)Br",
                functional_groups=stk.BromoFactory(deleters=(), bonders=(1,)),
            ),
            generator=bbprep.generators.SelectorDistanceScanner(
                selector=bbprep.selectors.BindersSelector(),
                scanned_changes=[-2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2.0],
            ),
            num_confs=9,
            min_energy=(10.901384927360974, 4),
            max_energy=(170.8986255707084, 8),
            energy_5=(10.901384927360974, 4),
            name=name,
        ),
    )
)
def molecule(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import numpy as np
import stk

from .modifier import Modifier


class RandomFGs(Modifier):
    """Modify a building block."""

    def modify(  # type: ignore[override]
        self,
        building_block: stk.BuildingBlock,
        desired_functional_groups: int,
        seed: int = 1000,
    ) -> stk.BuildingBlock:
        selected_fgs: list[stk.FunctionalGroup]
        if (
            building_block.get_num_functional_groups()
            == desired_functional_groups
        ):
            return building_block.clone()

        if (
            building_block.get_num_functional_groups()
            < desired_functional_groups
        ):
            msg = (
                f"{building_block} has less functional groups than"
                f" asked for ({desired_functional_groups})."
            )
            raise RuntimeError(msg)

        generator = np.random.default_rng(seed)

        existing_fgs = list(building_block.get_functional_groups())

        selected_fgs = list(
            generator.choice(
                np.array(existing_fgs),
                size=desired_functional_groups,
                replace=False,
            )
        )

        return building_block.with_functional_groups(selected_fgs)
